
#include <sjimage2graph.h>
#include <math.h>
#include <stdint.h>
#include <sjkdtree.h>

void RGB2Lab(float R, float G, float B,
             int *L, int *a, int *b)
{
    float X, Y, Z, fX, fY, fZ;
    
    X = 0.412453*R + 0.357580*G + 0.180423*B;
    Y = 0.212671*R + 0.715160*G + 0.072169*B;
    Z = 0.019334*R + 0.119193*G + 0.950227*B;
    
    X /= (255 * 0.950456);
    Y /=  255;
    Z /= (255 * 1.088754);
    
    if (Y > 0.008856)
    {
        fY = pow(Y, 1.0/3.0);
        *L = (int)(116.0*fY - 16.0 + 0.5);
    }
    else
    {
        fY = 7.787*Y + 16.0/116.0;
        *L = (int)(903.3*Y + 0.5);
    }
    
    if (X > 0.008856)
        fX = pow(X, 1.0/3.0);
    else
        fX = 7.787*X + 16.0/116.0;
    
    if (Z > 0.008856)
        fZ = pow(Z, 1.0/3.0);
    else
        fZ = 7.787*Z + 16.0/116.0;
    
    *a = (int)(500.0*(fX - fY) + 0.5);
    *b = (int)(200.0*(fY - fZ) + 0.5);
    
    //printf("RGB=(%d,%d,%d) ==> Lab(%d,%d,%d)\n",R,G,B,*L,*a,*b);
}



#define MAXVOISIN 50
#define SAVEREVERSEMAP
static double dist_sq( double *a1, double *a2, int dims ) {
  double dist_sq = 0, diff;
  while( --dims >= 0 ) {
    diff = (a1[dims] - a2[dims]);
    dist_sq += diff*diff;
  }
  return dist_sq;
}

static double rd( void ) {
  return (double)rand()/RAND_MAX * 20.0;// - 10.0;
}

double * rdV( void ) {
  double * v = (double*) malloc (sizeof(double)*2);//new double[2];
  v[0] = (double)rand()/RAND_MAX * 20.0;;
  v[1] = (double)rand()/RAND_MAX * 20.0;;

  return v;
  //return (double)rand()/RAND_MAX * 20.0;// - 10.0;
}


int32_t sjppm2graph(struct xvimage *r, struct xvimage *v, struct xvimage *b, int32_t param, graphe ** gA, double ** FvA)
{
  int32_t i,j;                                /* index muet */
  int32_t rs = rowsize(r);                   /* taille ligne */ 
  int32_t cs = colsize(r);                   /* taille colone */
  int32_t N = rs * cs;                        /* taille image */
  uint8_t *R = UCHARDATA(r);        /* composante rouge */
  uint8_t *V = UCHARDATA(v);        /* composante verte */
  uint8_t *B = UCHARDATA(b);        /* composante bleue */
  graphe *g;
  double valor = 0;
  int32_t x,y;

  uint8_t *F;
  double *Fv;
  int lx = 0, ly = 0;
  int ax = 0, ay = 0;
  int bx = 0, by =0;

    int lx2 = 0, ly2 = 0;
    int ax2 = 0, ay2 = 0;
    int bx2 = 0, by2 =0;

    
  g = initgraphe(N, 2*(2*N-rs-cs));

  switch(param)
  {
  case 0: /*MAX OF THREE CHANNELS*/

    setSize(g,rs,cs);
    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      {
	if (i < rs - 1)
	{
	  x = j * rs + i;
	  y = j * rs + i + 1;
		valor = (mcmax(mcabs( (int32_t)(B[x]) - (int32_t)(B[y]) ),
			       mcmax(mcabs( (int32_t)(R[x]) - (int32_t)(R[y]) ),
			             mcabs( (int32_t)(V[x]) - (int32_t)(V[y])))));

	  addarete(g, x, y, (double)(valor));
	}
	if (j < cs - 1)
	{
	  x = j * rs + i;
	  y = (j+1) * rs + i;
		valor = (mcmax(mcabs( (int32_t)(B[x]) - (int32_t)(B[y]) ),
			       mcmax(mcabs( (int32_t)(R[x]) - (int32_t)(R[y]) ),
			             mcabs( (int32_t)(V[x]) - (int32_t)(V[y])))));

	  addarete(g, x, y, (double)(valor));
	}
      }


  break;
  case 1: /*EUCLIDEAN DISTANCE*/
	      setSize(g,rs,cs);
    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      {
	if (i < rs - 1)
	{
	  x = j * rs + i;
	  y = j * rs + i + 1;
		valor = ( sqrt ( 
					   ( (double)(B[x]) - (double)(B[y])) * ( (double)(B[x]) - (double)(B[y])) +  
					   ( (double)(R[x]) - (double)(R[y])) * ( (double)(R[x]) - (double)(R[y])) +  
					   ( (double)(V[x]) - (double)(V[y])) * ( (double)(V[x]) - (double)(V[y]))
					   ));

	  addarete(g, x, y, (double)(valor));
	}
	if (j < cs - 1)
	{
	  x = j * rs + i;
	  y = (j+1) * rs + i;
		valor = (sqrt ( 
					   ( (double)(B[x]) - (double)(B[y])) * ( (double)(B[x]) - (double)(B[y])) +  
					   ( (double)(R[x]) - (double)(R[y])) * ( (double)(R[x]) - (double)(R[y])) +  
					   ( (double)(V[x]) - (double)(V[y])) * ( (double)(V[x]) - (double)(V[y]))
					   ));

	  addarete(g, x, y, (double)(valor));
	}
      }

   break;
      case 2: /*EUCLIDEAN DISTANCE USING LAB SPACE*/
              /*RGB => LAB >= DISTANCE*/
	      setSize(g,rs,cs);
          for (j = 0; j < cs; j++)
              for (i = 0; i < rs; i++)
              {
                  if (i < rs - 1)
                  {
                      x = j * rs + i;
                      y = j * rs + i + 1;
                      lx = 0; ly = 0;
                      ax = 0; ay = 0;
                      bx = 0; by = 0;
                      RGB2Lab(R[x],V[x],B[x],&lx,&ax,&bx);
                      RGB2Lab(R[y],V[y],B[y],&ly,&ay,&by);
                      valor = ( sqrt (
                                      ( (double)(lx) - (double)(ly)) * ( (double)(lx) - (double)(ly)) +
                                      ( (double)(ax) - (double)(ay)) * ( (double)(ax) - (double)(ay)) +
                                      ( (double)(bx) - (double)(by)) * ( (double)(bx) - (double)(by))
                                      ));
                      
                      addarete(g, x, y, (double)(valor));
                  }
                  if (j < cs - 1)
                  {
                      x = j * rs + i;
                      y = (j+1) * rs + i;
                      lx = 0; ly = 0;
                      ax = 0; ay = 0;
                      bx = 0; by = 0;
                      RGB2Lab(R[x],V[x],B[x],&lx,&ax,&bx);
                      RGB2Lab(R[y],V[y],B[y],&ly,&ay,&by);
                      valor = ( sqrt (
                                      ( (double)(lx) - (double)(ly)) * ( (double)(lx) - (double)(ly)) +
                                      ( (double)(ax) - (double)(ay)) * ( (double)(ax) - (double)(ay)) +
                                      ( (double)(bx) - (double)(by)) * ( (double)(bx) - (double)(by))
                                      ));
                      
                      addarete(g, x, y, (double)(valor));
                  }
              }
          
          break;
      case 3: /*EUCLIDEAN DISTANCE USING LAB SPACE - NEGATIVE IMAGE*/
          /*RGB => NEGATIVE => LAB >= DISTANCE*/
	      setSize(g,rs,cs);
          for (j = 0; j < cs; j++)
              for (i = 0; i < rs; i++)
              {
                  if (i < rs - 1)
                  {
                      x = j * rs + i;
                      y = j * rs + i + 1;
                      lx = 0; ly = 0;
                      ax = 0; ay = 0;
                      bx = 0; by = 0;
                      RGB2Lab(255-R[x],255-V[x],255-B[x],&lx,&ax,&bx);
                      RGB2Lab(255-R[y],255-V[y],255-B[y],&ly,&ay,&by);
                      valor = ( sqrt (
                                      ( (double)(lx) - (double)(ly)) * ( (double)(lx) - (double)(ly)) +
                                      ( (double)(ax) - (double)(ay)) * ( (double)(ax) - (double)(ay)) +
                                      ( (double)(bx) - (double)(by)) * ( (double)(bx) - (double)(by))
                                      ));
                      
                      addarete(g, x, y, (double)(valor));
                  }
                  if (j < cs - 1)
                  {
                      x = j * rs + i;
                      y = (j+1) * rs + i;
                      lx = 0; ly = 0;
                      ax = 0; ay = 0;
                      bx = 0; by = 0;
                      RGB2Lab(255-R[x],255-V[x],255-B[x],&lx,&ax,&bx);
                      RGB2Lab(255-R[y],255-V[y],255-B[y],&ly,&ay,&by);
                      valor = ( sqrt (
                                      ( (double)(lx) - (double)(ly)) * ( (double)(lx) - (double)(ly)) +
                                      ( (double)(ax) - (double)(ay)) * ( (double)(ax) - (double)(ay)) +
                                      ( (double)(bx) - (double)(by)) * ( (double)(bx) - (double)(by))
                                      ));
                      
                      addarete(g, x, y, (double)(valor));
                  }
              }
          
          break;
      case 4: /*EUCLIDEAN DISTANCE USING LAB SPACE - MAX(ORIGINAL, NEGATIVE) IMAGE*/
          /*RGB => NEGATIVE => LAB >= DISTANCE*/
	      setSize(g,rs,cs);
          for (j = 0; j < cs; j++)
              for (i = 0; i < rs; i++)
              {
                  if (i < rs - 1)
                  {
                      x = j * rs + i;
                      y = j * rs + i + 1;
                      lx = 0; ly = 0;
                      ax = 0; ay = 0;
                      bx = 0; by = 0;
                      RGB2Lab(R[x],V[x],B[x],&lx,&ax,&bx);
                      RGB2Lab(R[y],V[y],B[y],&ly,&ay,&by);
                      RGB2Lab(255-R[x],255-V[x],255-B[x],&lx2,&ax2,&bx2);
                      RGB2Lab(255-R[y],255-V[y],255-B[y],&ly2,&ay2,&by2);
                      lx = mcmax(lx,lx2); ly = mcmax(ly,ly2);
                      ax = mcmax(ax,ax2); ay = mcmax(ay,ay2);
                      bx = mcmax(bx,bx2); by = mcmax(by,by2);
                      
                      valor = ( sqrt (
                                      ( (double)(lx) - (double)(ly)) * ( (double)(lx) - (double)(ly)) +
                                      ( (double)(ax) - (double)(ay)) * ( (double)(ax) - (double)(ay)) +
                                      ( (double)(bx) - (double)(by)) * ( (double)(bx) - (double)(by))
                                      ));
                      
                      addarete(g, x, y, (double)(valor));
                  }
                  if (j < cs - 1)
                  {
                      x = j * rs + i;
                      y = (j+1) * rs + i;
                      lx = 0; ly = 0;
                      ax = 0; ay = 0;
                      bx = 0; by = 0;
                      RGB2Lab(R[x],V[x],B[x],&lx,&ax,&bx);
                      RGB2Lab(R[y],V[y],B[y],&ly,&ay,&by);
                      RGB2Lab(255-R[x],255-V[x],255-B[x],&lx2,&ax2,&bx2);
                      RGB2Lab(255-R[y],255-V[y],255-B[y],&ly2,&ay2,&by2);
                      lx = mcmax(lx,lx2); ly = mcmax(ly,ly2);
                      ax = mcmax(ax,ax2); ay = mcmax(ay,ay2);
                      bx = mcmax(bx,bx2); by = mcmax(by,by2);
                      
                      
                      valor = ( sqrt (
                                      ( (double)(lx) - (double)(ly)) * ( (double)(lx) - (double)(ly)) +
                                      ( (double)(ax) - (double)(ay)) * ( (double)(ax) - (double)(ay)) +
                                      ( (double)(bx) - (double)(by)) * ( (double)(bx) - (double)(by))
                                      ));
                      
                      addarete(g, x, y, (double)(valor));
                  }
              }
          
          break;

          
	case 6: /*GRADIENT ALREADY COMPUTED*/

    setSize(g,rs,cs);
    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      {
	if (i < rs - 1)
	{
	  x = j * rs + i;
	  y = j * rs + i + 1;
	valor = (mcmax(mcmax( (int32_t)(B[x]),(int32_t)(B[y]) ),
  	       mcmax(mcmax( (int32_t)(R[x]),(int32_t)(R[y]) ),
           mcmax( (int32_t)(V[x]),(int32_t)(V[y])))));


	  addarete(g, x, y, (double)(valor));
	}
	if (j < cs - 1)
	{
	  x = j * rs + i;
	  y = (j+1) * rs + i;
	valor = (mcmax(mcmax( (int32_t)(B[x]),(int32_t)(B[y]) ),
  	       mcmax(mcmax( (int32_t)(R[x]),(int32_t)(R[y]) ),
           mcmax( (int32_t)(V[x]),(int32_t)(V[y])))));


	  addarete(g, x, y, (double)(valor));

	}
      }


  break;

  }

    Fv = (double*) calloc(N, sizeof(double));
    if(Fv == NULL){
      fprintf(stderr,"cannot allocate (%d bytes) memory\n", N*sizeof(double));
      terminegraphe(g);
      exit(1);
    }
  
    for(i = 0; i < N; i++){
      Fv[i] = 1.0;
    }

	*FvA = Fv;
    *gA = g;

	return 1;
}

int32_t sjpgm2graph(struct xvimage *Im, int32_t mode, graphe ** gA, double ** FvA){

  graphe *g;
  int32_t rs, cs, ds, N, i,j,x,y;
  uint8_t *F;
  double *Fv; 
  char  nome[80];
  
  if(depth(Im) != 1) {
    fprintf(stderr, " 3D not yet implemented\n");
    exit(1);
  } 
  rs = rowsize(Im);
  cs = colsize(Im);
  N = rs*cs;

  switch(datatype(Im)){
  case 	VFF_TYP_1_BYTE: 
    F = UCHARDATA(Im);
    break;
  default : 
    fprintf(stderr,"%s: not yet implemented for this kind of data - only byte images implemented \n");
    exit(0);
  }
  
  switch(mode){
  case 0:
    /* Case of a 4-connected graph where each edge is weighted by the
       absolute difference of intensity between its extremity
       pixels */
    g = initgraphe(N, 2*(2*N-rs-cs));
    setSize(g,rs,cs);
    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      {
	if (i < rs - 1)
	{
	  x = j * rs + i;
	  y = j * rs + i + 1;
	  addarete(g, x, y, (double)(abs( (int)F[x] - (int)F[y])));
	}
	if (j < cs - 1)
	{
	  x = j * rs + i;
	  y = (j+1) * rs + i;
	  addarete(g, x, y, (double)(abs( (int)F[x] - (int)F[y])));
	}
      }
    
    Fv = (double*) calloc(N, sizeof(double));
    if(Fv == NULL){
      fprintf(stderr," cannot allocate (%d bytes) memory\n", N*sizeof(double));
      terminegraphe(g);
      exit(1);
    }
    /* The area of any pixel is 1 */

    for(i = 0; i < N; i++){
      Fv[i] = 1.0;
    }


//    SaveGraphe(g, argv[3], Fv);
//    terminegraphe(g);
//    free(Fv);
    break;
  case 6:
    /* Case of a 4-connected graph where each edge is weighted by the
       absolute difference of intensity between its extremity
       pixels */
    g = initgraphe(N, 2*(2*N-rs-cs));
    setSize(g,rs,cs);
    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      {
	if (i < rs - 1)
	{
	  x = j * rs + i;
	  y = j * rs + i + 1;
	  addarete(g, x, y, (double)(max( (int)F[x],(int)F[y])));
	}
	if (j < cs - 1)
	{
	  x = j * rs + i;
	  y = (j+1) * rs + i;
	  addarete(g, x, y, (double)(max( (int)F[x],(int)F[y])));
	}
      }
    
    Fv = (double*) calloc(N, sizeof(double));
    if(Fv == NULL){
      fprintf(stderr," cannot allocate (%d bytes) memory\n", N*sizeof(double));
      terminegraphe(g);
      exit(1);
    }
    /* The area of any pixel is 1 */

    for(i = 0; i < N; i++){
      Fv[i] = 1.0;
    }

	break;
	default :
    fprintf(stderr, "mode %d not available\n", mode);
    exit(0);
  }
//  freeimage(Im);
	*FvA = Fv;
    *gA = g;
	return 1;
}

int32_t sjpgm2knngraph(struct xvimage *Im, int32_t mode, graphe ** gA, double ** FvA, double radius){

  int i, num_pts = 0;
  void  *ptree;
  char *data, *pch;
  struct kdres *presults;
  double pos[2], dist;
  double pt[2] = { 0, 1 };
//  double radius = 1;
  double * v = NULL;
  int ndim = 2;
  graphe *g;
  int32_t rs, cs, ds, N, j,x,y;
  uint8_t *F;
  double *Fv; 
  char  nome[80];
  int k = 0;

  if(depth(Im) != 1) {
    fprintf(stderr, " 3D not yet implemented\n");
    exit(1);
  } 
  rs = rowsize(Im);
  cs = colsize(Im);
  N = rs*cs;

  switch(datatype(Im)){
  case 	VFF_TYP_1_BYTE: 
    F = UCHARDATA(Im);
    break;
  default : 
    fprintf(stderr,"%s: not yet implemented for this kind of data - only byte images implemented \n");
    exit(0);
  }
  num_pts = N;

  if(!(data = (char*) malloc(num_pts))) {
    perror("malloc failed");
    return 1;
  }

  /* create a k-d tree for 3-dimensional points */
  ptree = kd_create( ndim );

    switch(mode){
  case 0:
    /* Case of a 4-connected graph where each edge is weighted by the
       absolute difference of intensity between its extremity
       pixels */
    g = sjinitgraphe(N,MAXVOISIN*N);//N*N );//2*(2*N-rs-cs));
    setSize(g,rs,cs);
    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      { 
          kd_insert2( (struct kdtree*) ptree, (double)j,(double)i, &k );// );
		  k++;
      }
    
    Fv = (double*) calloc(N, sizeof(double));
    if(Fv == NULL){
      fprintf(stderr,"cannot allocate (%d bytes) memory\n", N*sizeof(double));
      terminegraphe(g);
      exit(1);
    }
    /* The area of any pixel is 1 */

    for(i = 0; i < N; i++){
      Fv[i] = 1.0;
    }

    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      { 

		pt[0] = j;
		pt[1] = i;
		/* find points closest to the origin and within distance radius */
        presults = kd_nearest_range(  (struct kdtree*) ptree, pt, radius );

		/* print out all the points found in results */
//		printf( "found %d results:\n", kd_res_size(presults) );

		  while( !kd_res_end( presults ) ) {
			/* get the data and position of the current result item */
			pch = (char*)kd_res_item( presults, pos );
    
			/* compute the distance of the current result from the pt */
			dist = sqrt( dist_sq( pt, pos, 2 ) );

			/* print out the retrieved data */
//			printf( "node at (%.3f, %.3f) is %.3f away and has data=%c\n", 
//				pos[0], pos[1],  dist, *pch );

			x = j * rs + i;
			y = pos[0] * rs + pos[1];
			if (x > y)
				addarete(g, x, y, (double)(abs( (int)F[x] - (int)F[y])));

			/* go to the next entry */
			kd_res_next( presults );
		  }
	  }
//    SaveGraphe(g, argv[3], Fv);
//    terminegraphe(g);
//    free(Fv);
    break;
  default :
    fprintf(stderr, "mode %d not available\n", mode);
    exit(0);
  }



  /* free our tree, results set, and other allocated memory */
  free( data );
  kd_res_free( presults );
  kd_free(  (struct kdtree*) ptree );

//  freeimage(Im);

    for(i = 0; i < N; i++){
      Fv[i] = 1.0;
    }

	*FvA = Fv;
    *gA = g;

	return 1;

}
int32_t sjppm2knngraph(struct xvimage *r, struct xvimage *v, struct xvimage *b, int32_t param, graphe ** gA, double ** FvA, double radius){

  int  num_pts = 0;
  void  *ptree;
  char *data, *pch;
  struct kdres *presults;
  double pos[2], dist;
  double pt[2] = { 0, 1 };
//  double radius = 1;
//  double * v = NULL;
  int ndim = 2;
  graphe *g;
//  int32_t rs, cs, ds, N, j,x,y;
  uint8_t *F;
  double *Fv; 
  char  nome[80];
  int k = 0;
  int mode = param;

  int32_t i,j;                                /* index muet */
  int32_t rs = rowsize(r);                   /* taille ligne */ 
  int32_t cs = colsize(r);                   /* taille colone */
  int32_t N = rs * cs;                        /* taille image */
  uint8_t *R = UCHARDATA(r);        /* composante rouge */
  uint8_t *V = UCHARDATA(v);        /* composante verte */
  uint8_t *B = UCHARDATA(b);        /* composante bleue */
  uint8_t valor = 0;
  int32_t x,y;


  g = initgraphe(N, 2*(2*N-rs-cs));

  setSize(g,rs,cs);
  rs = rowsize(r);
  cs = colsize(r);
  N = rs*cs;
  num_pts = N;

  if(!(data = (char*) malloc(num_pts))) {
    perror("malloc failed");
    return 1;
  }

  /* create a k-d tree for 3-dimensional points */
  ptree = kd_create( ndim );

    switch(mode){
  case 0:
    /* Case of a 4-connected graph where each edge is weighted by the
       absolute difference of intensity between its extremity
       pixels */
    g = initgraphe(N, MAXVOISIN*N);// N*N );//2*(2*N-rs-cs));
    setSize(g,rs,cs);
    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      { 
          kd_insert2( (struct kdtree*) ptree, (double)j,(double)i, &k );// );
		  k++;
      }
    
    Fv = (double*) calloc(N, sizeof(double));
    if(Fv == NULL){
      fprintf(stderr,"cannot allocate (%d bytes) memory\n", N*sizeof(double));
      terminegraphe(g);
      exit(1);
    }
    /* The area of any pixel is 1 */

    for(i = 0; i < N; i++){
      Fv[i] = 1.0;
    }

    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++)
      { 

		pt[0] = j;
		pt[1] = i;
		/* find points closest to the origin and within distance radius */
        presults = kd_nearest_range(  (struct kdtree*) ptree, pt, radius );

		/* print out all the points found in results */
//		printf( "found %d results:\n", kd_res_size(presults) );

		  while( !kd_res_end( presults ) ) {
			/* get the data and position of the current result item */
			pch = (char*)kd_res_item( presults, pos );
    
			/* compute the distance of the current result from the pt */
			dist = sqrt( dist_sq( pt, pos, 2 ) );

			/* print out the retrieved data */
//			printf( "node at (%.3f, %.3f) is %.3f away and has data=%c\n", 
//				pos[0], pos[1],  dist, *pch );

			x = j * rs + i;
			y = pos[0] * rs + pos[1];
			if (x > y){
			
				valor = (uint8_t)(mcmax(mcabs( (int32_t)(B[j*rs+i]) - (int32_t)(B[j*rs+i+1]) ),
								mcmax(mcabs( (int32_t)(R[j*rs+i]) - (int32_t)(R[j*rs+i+1]) ),
				    			mcabs( (int32_t)(V[j*rs+i]) - (int32_t)(V[j*rs+i+1])))));
				 addarete(g, x, y, (double)(valor));

			}
			/* go to the next entry */
			kd_res_next( presults );
		  }
	  }
//    SaveGraphe(g, argv[3], Fv);
//    terminegraphe(g);
//    free(Fv);
    break;
  default :
    fprintf(stderr, "mode %d not available\n", mode);
    exit(0);
  }



  /* free our tree, results set, and other allocated memory */
  free( data );
  kd_res_free( presults );
  kd_free(  (struct kdtree*) ptree );

//  freeimage(Im);
    Fv = (double*) calloc(N, sizeof(double));
    if(Fv == NULL){
      fprintf(stderr,"cannot allocate (%d bytes) memory\n", N*sizeof(double));
      terminegraphe(g);
      exit(1);
    }
  
    for(i = 0; i < N; i++){
      Fv[i] = 1.0;
    }

	*FvA = Fv;
    *gA = g;

	return 1;

  return 1;

}

int32_t sjrgb2knngraph(struct xvimage *r, struct xvimage *v, struct xvimage *b, int32_t param, graphe ** gA, double ** FvA, double radius){
	int i, j;
	int * hist;
	int * reverse;

  int  num_pts = 0;
  void  *ptree;
  char *data, *pch;
  struct kdres *presults;
  int ndim = 3;
  double pos[3], dist;
  double pt[3] = { 0, 0, 0 };

	graphe *g;
  uint8_t *F;
  double *Fv; 
  int k = 0;
  int mode = param;

  int32_t rs = rowsize(r);                   /* taille ligne */ 
  int32_t cs = colsize(r);                   /* taille colone */
  int32_t N = rs * cs;                        /* taille image */
  uint8_t *R = UCHARDATA(r);        /* composante rouge */
  uint8_t *V = UCHARDATA(v);        /* composante verte */
  uint8_t *B = UCHARDATA(b);        /* composante bleue */
  uint8_t valor = 0;
  int32_t x,y;
  uint8_t red, green, blue;
	int l = 0;
	int tam = 0;
	int ALT = 256, LARG = 256, PROF = 256;
//	int k = 0;
#ifdef SAVEREVERSEMAP
  FILE * file = fopen("lixo.txt","w");
#endif

	hist = (int*) malloc (sizeof(int)*256*256*256);
	reverse = (int*) malloc (sizeof(int)*256*256*256);
  
	if(!(data = (char*) malloc(256*256*256))) {
    perror("malloc failed");
    return 1;
  }

  ptree = kd_create( ndim );

 	for (i = 0; i < 256; i++)
		for (j = 0; j < 256; j++)
			for (l = 0; l < 256; l++){
				x = i *LARG + j + ALT*LARG*l; 
				hist[x] = 0;
				reverse[x] = -1;
			}
				

    for (j = 0; j < cs; j++)
      for (i = 0; i < rs; i++){
				x = j * rs + i;
				red = R[x];
				green = V[x];
				blue = B[x];
				
				y = red *LARG + green + ALT*LARG*blue; 

				if (hist[y] == 0) tam++;
				hist[y] =  hist[y]+1;
				
			}

//		g = initgraphe(ALT*LARG*PROF, MAXVOISIN*tam);
		g = initgraphe(tam, MAXVOISIN*tam);
		k = 0;
		for (i = 0; i < 256; i++)
			for (j = 0; j < 256; j++)
				for (l = 0; l < 256; l++){
					x = i *LARG + j + ALT*LARG*l; 
					if (hist[x] != 0){
						reverse[x] = k;	
#ifdef SAVEREVERSEMAP
						fprintf(file,"%d %d %d %d\n", k,i,j,l);//FILE * file = fopen("lixo.txt","w");
#endif
                        kd_insert3( (struct kdtree*) ptree, (double)i,(double)j,(double)l, &k );
						k++;
					}
				}
#ifdef SAVEREVERSEMAP
						fclose(file);
#endif   

	for (i = 0; i < 256; i++)
		for (j = 0; j < 256; j++)
			for (l = 0; l < 256; l++){
					x = i *LARG + j + ALT*LARG*l; 
					if (hist[x] != 0){

						pt[0] = i;
						pt[1] = j;
						pt[2] = l;
                        presults = kd_nearest_range(  (struct kdtree*) ptree, pt, radius );
						while( !kd_res_end( presults ) ) {
							pch = (char*)kd_res_item( presults, pos );
							x = i *LARG + j + ALT*LARG*l; 
							y = pos[0] *LARG + pos[1] + ALT*LARG*pos[2]; 
		          valor = sqrt(mcabs(i-pos[0]) * mcabs(i-pos[0]) +
								mcabs(j-pos[1]) * mcabs(j-pos[1]) +
								mcabs(l-pos[2]) * mcabs(l-pos[2]));
							if (x > y){	
								addarete(g, reverse[x], reverse[y], (double)(valor));
							}
							kd_res_next( presults );
						}
					}
			}
  /* free our tree, results set, and other allocated memory */
  free( data );
  kd_res_free( presults );
  kd_free(  (struct kdtree*) ptree );

//	N = ALT*PROF*LARG;
	N = tam;
  Fv = (double*) calloc(N, sizeof(double));
  if(Fv == NULL){
    fprintf(stderr,"cannot allocate (%d bytes) memory\n", N*sizeof(double));
    terminegraphe(g);
    exit(1);
  }
  for(i = 0; i < N; i++) Fv[i] = 1.0;
	*FvA = Fv;
  *gA = g;
	free(hist);
	return 1;
}
#ifdef TEST
int main()
{
  struct xvimage *im = readimage("test1.pgm");
  struct xvimage *am;
  if (!im) exit(1);
  am = copyimage(im);
  printimage(am);
  writeimage(am, "test2.pgm");
  razimage(am);
  copy2image(im, am);
  writeimage(am, "test3.pgm");
}
#endif
