%Author: Benjamin Perret ESIEE Paris 2017

function [ucm,time] = gradient2WSHierarchy(gradient)
 
    N=numel(gradient);
    [h,w]=size(gradient);
    
    %% create 4 adjacency edges
    g = zeros(size(gradient)+1);
    g(1:end-1,1:end-1) = gradient; % add extra edges for the ease of indexing
    hE = (g(1:end-1,1:end-1)+g(1:end-1,2:end))/2;
    vE = (g(1:end-1,1:end-1)+g(2:end,1:end-1))/2;
    
    % packing
    hE = reshape(transpose(hE),1,N);
    vE = reshape(transpose(vE),1,N);
    edges = uint8([hE, vE]);

    %% compute saliency map (UCM)
    tic
    [sm, time] = WSHierarchy(edges, h, w, 0);
    fprintf('Tic toc time %2.1f ms, measured inside function %d ms\n',toc()*1000,time);
    
    %% unpack edges and embed in 2D khalimsky grid
    ucm = zeros(size(gradient)*2-1);
    hE = transpose(reshape(sm(:,1:end/2),w,h));
    vE = transpose(reshape(sm(:,end/2+1:end),w,h));
    
    ucm(1:2:end,2:2:end-1) = hE(:,1:end-1);
    ucm(2:2:end-1,1:2:end) = vE(1:end-1,:);
    
    ucm(2:2:end-1,2:2:end-1) = max(max(ucm(2:2:end-1,1:2:end-2),...
                                       ucm(2:2:end-1,3:2:end)),...
                                   max(ucm(1:2:end-2,2:2:end-1),...
                                        ucm(3:2:end,2:2:end-1)));
end