#include <emscripten.h> 
#include <vector>
#include <cmath>
#include <algorithm>



typedef unsigned char uchar;

typedef float (*Gradient)(uchar, uchar, uchar, uchar, uchar, uchar);

using namespace std;


// extern fonction for ui print
extern "C" {
  extern void ui(const char[]);
}


#define ADJ_4 4
#define ADJ_8 8


// euclidean distance 
float distance(float r1, float g1, float b1, float r2, float g2, float b2)
{
    float dr=r1-r2 ;
    float dg=g1-g2 ;
    float db=b1-b2 ;
    return sqrt(dr*dr+dg*dg+db*db );
}

struct Image {
	uchar * data; // RGBA raster
	int height;
	int width;
};




struct Edge {
	int p1;
	int p2;
	float w;
	
	Edge()
	{

	}
	
	Edge(int _p1, int _p2, float _w)
	{
		p1 = _p1;
		p2 = _p2;
		w = _w;
	}
};


// for sorting edges
bool compFunc (Edge & i,Edge & j) { return (i.w<j.w); }

// binary tree only
struct Child{
	int c1;
	int c2;
};

// global vars
struct Context {
	Image image;

	/* Kruskal data */
	int * parentUF; // union find parent relation
	int * rankUF; // union find rank measure
	int * root; // subtree roots
	
	/* Kruskal result */
	Edge * mst;
	
	
	/* Tree */
	int * parent;
	int nbNodes;
	Child * children;
	
	/* attributs */
	float * area;
	float * volume;
	int * depth;
	float * level;
};

int getImageSize(Context & context)
{
	return context.image.height*context.image.width;
}


float * convertRGBtoLAB(Context & context)
{
	uchar * data= context.image.data;
	
	int size = getImageSize(context);
	
	float * res = new float[size*3];
	for(int i=0, j=0;i<size*4;i+=4, j+=3)
	{
		uchar r1 = data[i];
		uchar g1 = data[i+1];
		uchar b1 = data[i+2];
		
		float var_R1 = ( r1 / 255.f ) ;       //R from 0 to 255
		float var_G1 = ( g1 / 255.f ) ;       //G from 0 to 255
		float var_B1 = ( b1 / 255.f ) ;       //B from 0 to 255

		if ( var_R1 > 0.04045f ) 
			var_R1 = pow( ( var_R1 + 0.055f ) / 1.055f ,2.4f);
		else  var_R1 = var_R1 / 12.92f;
		if ( var_G1 > 0.04045f )
			var_G1 = pow( ( var_G1 + 0.055f ) / 1.055f,  2.4f);
		else  var_G1 = var_G1 / 12.92f;
		if ( var_B1 > 0.04045f ) 
			var_B1 = pow( ( var_B1 + 0.055f ) / 1.055f,  2.4f);
		else var_B1 = var_B1 / 12.92f;

		var_R1 = var_R1 * 100.f;
		var_G1 = var_G1 * 100.f;
		var_B1 = var_B1 * 100.f;

		//Observer. = 2°, Illuminant = D65
		float var_X1 = (var_R1 * 0.4124f + var_G1 * 0.3576f + var_B1 * 0.1805f) / 95.047f;          //ref_X =  95.047   Observer= 2°, Illuminant= D65
		float var_Y1 = (var_R1 * 0.2126f + var_G1 * 0.7152f + var_B1 * 0.0722f) / 100.000f;         //ref_Y = 100.000
		float var_Z1 = (var_R1 * 0.0193f + var_G1 * 0.1192f + var_B1 * 0.9505f) / 108.883f;         //ref_Z = 108.883

		if ( var_X1 > 0.008856f ) 
			var_X1 = pow( var_X1 , ( 1.f/3.f ));
		else var_X1 = ( 7.787f * var_X1 ) + ( 16.f / 116.f );
		if ( var_Y1 > 0.008856f ) 
			var_Y1 = pow( var_Y1 , ( 1./3.f ));
		else  var_Y1 = ( 7.787f * var_Y1 ) + ( 16.f / 116.f );
		if ( var_Z1 > 0.008856f ) 
			var_Z1 = pow( var_Z1 , ( 1.f/3.f ));
		else var_Z1 = ( 7.787f * var_Z1 ) + ( 16.f / 116.f );

		float L1 = ( 116.f * var_Y1 ) - 16.f;
		float A1 = 500.f * ( var_X1 - var_Y1 );
		float B1 = 200.f * ( var_Y1 - var_Z1 );
	
		res[j] = L1;
		res[j+1] = A1;
		res[j+2] = B1;
	}
	return res;
}

// [[0;255]] -> [0;1]
float * convertRGBtoRGB(Context & context)
{
	uchar * data= context.image.data;
	
	int size = getImageSize(context);
	
	float * res = new float[size*3];
	for(int i=0, j=0;i<size*4;i+=4, j+=3)
	{
		res[j] = data[i]/255.F;
		res[j+1] = data[i+1]/255.F;
		res[j+2] = data[i+2]/255.F;
	}
	return res;
}

void allocateGlobalMemory(Context & context){
	int size = context.image.width*context.image.height;
	context.nbNodes=size;
	
    context.parent = new int[2*size-1];

    context.children = new Child[2*size-1];
	
	context.parentUF = new int[size];
    context.rankUF = new int[size];
    context.root = new int[size];

	
	
}

void deleteGlobalMemory(Context & context)
{
	delete [] context.parentUF;
	delete [] context.rankUF;
	delete [] context.root;
}

// union find, construct getImageSize(context) singletons
void initUFStructures(Context & context) {

	int size=getImageSize(context);
	context.nbNodes=size;
	for (int i = 0; i < size; i++)
    {
        context.parentUF[i] = -1;
        context.rankUF[i] = 0;
        context.root[i] = i;
    }
	
    
}

// union find : find the canonical node of pos 
int findCanonical(int * parentUF, int pos)
{
    int t = pos;
    while ((parentUF[t]) != - 1)
        t = parentUF[t];
    while ((parentUF[pos]) != -1)
    {
        int tmp = parentUF[pos] ;
        parentUF[pos] = t  ;
        pos = tmp ;
    }
    return t ;
}


// union find : link canonical nodes c1 and c2
int link(int c1, int c2, Context & context )
{
    int r1 = context.root[c1];
    int r2 = context.root[c2];
	
	int size=context.nbNodes;
    // new node and rooting
    context.parent[r1] = size;
    context.parent[r2] = size;


    if ((context.rankUF[c1]) > (context.rankUF[c2])) {
        int tmp = c2 ;
        c2 = c1 ;
        c1 = tmp ;
    } else if (context.rankUF[c1] == context.rankUF[c2])
    {
        context.rankUF[c2]++;
    }
    context.parentUF[c1] = c2;
    context.root[c2] = size;
    context.children[size].c1=r1;
	context.children[size].c2=r2;
    context.parent[size]=-1;
	context.nbNodes++;
	return c2;
}


// explicit edges construction and weighting
Edge * constructAdjEdges(Image & image, float * data, int adj)
{
	int width = image.width;
	int height = image.height;
	
	int nbEdges=2*width*height-width-height;
	if(adj==ADJ_8)
	{
		nbEdges +=  2*(width-1)*(height-1);
	}
	Edge * edges = new Edge[nbEdges];
	int ne=0;
	float d;
	int pos1=0;
	 for (int y = 0; y < height; y++)
    {
        for (int x = 0; x < width; x++)
        {
            int bx = (x < width - 1);
            int by = (y < height - 1);
            int p4 = pos1 * 3;
            float p4r=data[p4];
            float p4g=data[p4+1];
            float p4b=data[p4+2];
            if (bx)
            {
                int pos2 = pos1 + 1;
                int pp4 = pos2 * 3;
                d = distance(p4r,p4g,p4b, data[pp4], data[pp4 + 1], data[pp4 + 2]);
				
                edges[ne++]=Edge(pos1, pos2, d);
            }
            if (by)
            {
                int pos2 = pos1 + width;
                int pp4 = pos2 * 3;
				
                d = distance(p4r,p4g,p4b, data[pp4], data[pp4 + 1], data[pp4 + 2]);
                edges[ne++]=Edge(pos1, pos2, d);
                if (adj == ADJ_8) {
                    if (bx)
                    {
                        pos2 = pos1 + width + 1;
                        int pp4 = pos2 * 3;
                        d = distance(p4r,p4g,p4b, data[pp4], data[pp4 + 1], data[pp4 + 2]);
                        edges[ne++]=Edge(pos1, pos2, d);
                    }
                    if (x > 0)
                    {
                        pos2 = pos1 + width - 1;
                        int pp4 = pos2 * 3;
                        d = distance(p4r,p4g,p4b, data[pp4], data[pp4 + 1], data[pp4 + 2]);
                        edges[ne++]=Edge(pos1, pos2, d);
                    }
                }
            }
            pos1++;
        }
    }
	ui("Sorting edges.");
	std::sort (edges, edges+ne, compFunc); 
	
	return edges;
}

// compute attributes : area, volume, depth, level
void finalizeTree(Context & context) {

	int size = getImageSize(context);
	int nbnodes = context.nbNodes; 
	Edge * mst = context.mst;
	int * parent = context.parent;
	
	float * area = new float[nbnodes];
	int * depth = new int[nbnodes];
	float * volume = new float[nbnodes];
	float * level = new float[nbnodes];
	
	context.area=area;
	context.depth=depth;
	context.volume=volume;
	context.level=level;
	
	for(int i =0;i<nbnodes;i++)
	{
		volume[i]=0.f;
		if(i<size)
		{
			area[i]=1;
			level[i]=-1.f;
		} else {
			area[i]=0;
			level[i]= mst[(i - size)].w;
		}
	}
	
	for(int i = 0; i<nbnodes; i++)
	{
		int p = parent[i];
		if(p!=-1)
		{
			area[p]+=area[i];
			if(i>=size)
			{
				float lvl = level[i];
				float lvlp = level[p];
				volume[i]+=(lvlp-lvl)*area[i];
			}
			volume[p]+=volume[i];
		}
	}
		
	depth[nbnodes-1]=0;
	for(int i=nbnodes-2;i>=0;i--)
	{
		int p = parent[i];
		 if (level[i] == level[p])
        {
            depth[i] = depth[p];
        } else {
            depth[i] = depth[p]+1;
        }
	}
}




// edges must be sorted
void constructBPT(Edge * edges, Context & context)
{
	int size = getImageSize(context);
	int nbEdges = size - 1;
    int nbeMST = 0;
	initUFStructures(context);
    Edge * mst = new Edge[nbEdges];
	
    for (int i = 0; nbeMST < nbEdges ; i++)
    {

        Edge e = edges[i];
        int c1 = findCanonical(context.parentUF, e.p1);
        int c2 = findCanonical(context.parentUF, e.p2);
        if (c1 != c2)
        {
            link(c1, c2, context);
            mst[nbeMST++]=e;
        }
    }
	
    context.mst = mst;
    finalizeTree(context);    
}


void destroyTree(Context & context)
{
	delete [] context.mst;
	delete [] context.area;
	delete [] context.volume;
	delete [] context.depth;
	delete [] context.level;
}

/**
Transform a BPT into a component tree
*/
void cleanTree(Context & context)
{

	int * parent = context.parent;
	float * level = context.level;
	int * depth = context.depth;
	
    int nbNodes = context.nbNodes;
	int nbLeaves = getImageSize(context);
    
	Child * children = context.children;
    int count = 0; // total number of dleted nodes
    uchar deleted [nbNodes]; // mark deleted nodes
    int map [nbNodes]; // number of deleted nodes before node i
    for(int i=0;i<nbLeaves;i++)
	{	
		deleted[i]=0;
		map[i]=0;
	}
	// from root to leaves, compute the new parent relation, don't care of the holes in the parent tab
    for (int i = nbNodes - 2; i >= nbLeaves; i--)
    {
        
        int p = parent[i];

        if (level[i] == level[p])
        {
            Child cl = children[i];
			parent[cl.c1]=p;
			parent[cl.c2]=p;
            deleted[i] = 1;
            count++;
        }else
		{
			deleted[i]=0;
		}
        map[i] = count; // inverse of what we want: number of deleted nodes after node i
    }
	//root
	map[nbNodes -1]=0;
	deleted[nbNodes -1]=0;
	
	 // correct the mapping
    for (int i = nbLeaves; i < nbNodes; i++)
    {
        map[i] = count - map[i];
    }
    
	
	// new relations with correct size
	int * nparent = new int[nbNodes-count];
	float * nlevel = new float[nbNodes-count];
	int *   ndepth = new int[nbNodes-count];

	
	count=0;
    for (int i = 0; i < nbNodes; i++)
    {
        if (deleted[i] != 1)
        {
            int p = parent[i];
            int np = (p != -1) ? (p - map[p]) : -1; // corrected parent index after holes removal
            
			nparent[count]=np;
			nlevel[count]=level[i];
			ndepth[count]=depth[i];
			count++;

        }
    }

   destroyTree(context);
   context.nbNodes=count;
   context.parent = nparent;
   context.level = nlevel;
   context.depth = ndepth;
    
}

// compute correct attribut values in the BPT
float * computeAttrMerged(Context & context, float * attribute)
{
	int * parent = context.parent;
	Child * children = context.children;
	float * level = context.level;
	int nbNodes = context.nbNodes;
    float * attr = new float[nbNodes];
	int nbLeaves = getImageSize(context);
    for (int i = 0; i < nbNodes; i++)
    {
        
		int p = parent[i];
        if (i<nbLeaves)
		{
			attr[i] = 0;
		}
		else if(p == -1 || level[i]!=level[p]) {
            attr[i] = attribute[i];
        } else {
			
            Child cl = children[i];
			float v1=attr[cl.c1];
			float v2=attr[cl.c2];
			if(v2 > v1)
				v1=v2;
            attr[i] = v1;
        }
    }
	return attr;

}

// reorder hierarchy w.r.t. a given attribute
Edge * reweightMST(Context & context, float * attr) {
	int size = getImageSize(context);
    Edge * mst = context.mst;
    Edge * nmst = new Edge[size-1];
	Child * children = context.children;

    int nbNodes = context.nbNodes;
    for (int i = size; i < nbNodes; i++)
    {
        float min = 99999999999;
        Child cl = children[i];
		float v1=attr[cl.c1];
		float v2=attr[cl.c2];
		if(v2 < v1)
			v1=v2;
        


        Edge e = mst[i - size];
        nmst[i - size]=Edge(e.p1,e.p2,v1);
    }
	return nmst;
}

// compute watershed, non mst edges are set to 0, mst edges remain unchanged
Edge * watershedMST(Context & context) {
	int size = getImageSize(context);
	int nbNodes = context.nbNodes;
    Edge * mst =context.mst;
    Edge * nmst = new Edge[size-1];
	int * parent = context.parent;
	Child * children = context.children;
	float * level = context.level;
	
    int * minima = new int[nbNodes];
	for(int i=0;i<size;i++)
		minima[i]=0;

  
    for (int i = size; i < nbNodes; i++)
    {
        Child cl = children[i];
		int m1 = minima[cl.c1];
		int m2 = minima[cl.c2];
		int nb = m1 + m2;
     
		Edge e =mst[i - size];
        float w = ((m1==0) || (m2==0)) ? 0 : e.w;
     
		nmst[i-size]=Edge(e.p1,e.p2,w);
       
        if (nb != 0)
        {
            minima[i] = nb;
        } else {
            int p = parent[i];
            if (p == -1 || (level[i] != level[p]))
            {
                minima[i] = 1;
               
            } else {
                minima[i] = 0;
            }
        }
    }
	
	delete [] minima;
	return nmst;
   

}

void filterHierarchyWatershed(Context & context) {

    ui("Finding watershed edges.");
    Edge * mst2 = watershedMST(context);
	
    ui("Constructing watershed hierarchy.");
	destroyTree(context);
	std::sort (mst2, mst2+getImageSize(context)-1, compFunc); 
    constructBPT(mst2, context);
	
	delete [] mst2;
	
}

void filterHierarchyAttribute(Context & context, float * attribute) {

	ui("Computing attributes.");
    float * attr = computeAttrMerged(context,attribute);
	
    ui("Simulating flooding.");
    Edge * mst2 = reweightMST(context,attr);
	
    ui("Constructing flooding hierarchy.");
	destroyTree(context);
	std::sort (mst2, mst2+getImageSize(context)-1, compFunc); 
    constructBPT(mst2, context);
	
	delete [] mst2;
	delete [] attr;
	
}

// least common ancestor, may be slow in badly balanced tree
int LCAnaif(int * parent, int * depth, int i, int j)
{
    while (i != j)
    {
        if (depth[i] > depth[j]) {
            i = parent[i];
        }
        else {
            j = parent[j];
        }
    }
    return i;
}

// compute pseudo saliency map for visualisation
uchar * saliency(Context & context, uchar * im2) //width, height, tree)
{
	int width = context.image.width;
	int height = context.image.height;
	int * parent = context.parent;
	int * depth = context.depth;
	float * level = context.level;

    int w = width * 2 - 1;
    int h = height * 2 - 1;
	
	// double resolution temp image
	float * im = new float [w * h];
	
  
 
	
    float vmin = 99999999999;
    float vmax = -1;
    
	// fill in the "correct" saliency map
    for (int y = 0; y < h; y++)
    {
        for (int x = 0; x < w; x++)
        {
            
            if ((x % 2 == 0) && (y % 2 == 0))
                im[y * w + x] = 0;
            else if ((x % 2 == 1) && (y % 2 == 0))
            {
                int p1 = (y >> 1) * width + ((x - 1) >>1);
                int p2 = p1 + 1;
                int l = LCAnaif(parent, depth, p1, p2);
                float d = level[l];
                im[y * w + x] = d;
                if (d < vmin)
                    vmin = d;
                if (d > vmax)
                    vmax = d;

            } else if ((x % 2 == 0) && (y % 2 == 1))
            {
                int p1 = ((y - 1) >>1) * width + (x >>1);
                int p2 = p1 + width;
                int l = LCAnaif(parent, depth, p1, p2);
                float d = level[l];
                im[y * w + x] = d;
                if (d < vmin)
                    vmin = d;
                if (d > vmax)
                    vmax = d;
            }
        }
    }
    
	// contrast adjustment
    for (int y = 0; y < h; y++)
    {
        for (int x = 0; x < w; x++)
        {
            if (((x % 2 == 1) && (y % 2 == 0)) || ((x % 2 == 0) && (y % 2 == 1)))
            {
                int pos = y * w + x;
                float v =  floor(pow((im[pos]-vmin) / (vmax-vmin),0.33)*255); // magic number !
                
                im[pos] = v;
            }
        }
    }
    int c=0;
	// final combination
    for (int y = 1; y < h; y += 2)
    {
        for (int x = 1; x < w; x += 2)
        {
            float v1 = im[(y - 1) * w + x];
            float v2 = im[(y) * w + x - 1];
            float v3 = im[(y) * w + x + 1];
            float v4 = im[(y + 1) * w + x];
            float v= fmax(0,fmin(255,fmax(v1,fmax(v2,fmax(v3,v4)))));
            
            
            im2[c++]=0;//255-v;
            im2[c++]=0;//255-v;
            im2[c++]=0;//255-v;
            im2[c++]=(uchar)v;
            
        }
    }

	delete [] im;
    return im2;
}

#define NO_ATTRIBUTE 0
#define AREA_ATTRIBUTE 1
#define VOLUME_ATTRIBUTE 2

#define GRAD_RGB 0
#define GRAD_LAB 1


extern "C" {
/**
 input:
	width
	height
	data: pixel raster, must be RGBA
	adj: neighbourhood : ADJ_4 or ADJ_8
	attributFilter: reordering according to an attribut : NO_ATTRIBUTE, AREA_ATTRIBUTE or VOLUME_ATTRIBUTE
	watershed: use watershed superpixels ? boolean
	gradient: edge weighting function : GRAD_RGB or GRAD_LAB
 output:
	parent: the parent relation (allocated by the function)
	nbnodes: number of elements in *parent
*/
void EMSCRIPTEN_KEEPALIVE constructHierarchy(int width, int height, uchar * data, int ** parent, int * nbnodes, int adj, int attributFilter, int watershed, int gradient) {

	Image image;
	image.data=data;
	image.height=height;
	image.width=width;
	Context context;
	context.image=image;
   
	
    ui("Allocating buffers.");
	allocateGlobalMemory(context);
	
	
	
	
		
	ui("Computing gradient.");
	float * transFormImage;
    Gradient grad;
	switch(gradient){
			case GRAD_RGB:
				transFormImage=convertRGBtoRGB(context);
				break;
			case GRAD_LAB:
				transFormImage=convertRGBtoLAB(context);
				break;
			default:
			EM_ASM({console.log('Incorrect gradient function: using LAB');}	);
			transFormImage=convertRGBtoLAB(context);
		}
	Edge * edges = constructAdjEdges(image, transFormImage, adj);
    delete transFormImage;

	ui("Constructing binary partition tree.");
	constructBPT(edges, context);
	delete [] edges;
	
	//ui("Measuring bassin sizes.");

	if(watershed!=0)
		filterHierarchyWatershed(context);
	
	if(attributFilter!=NO_ATTRIBUTE)
	{
		float * attribute; 
		switch(attributFilter){
			case AREA_ATTRIBUTE:
				attribute = context.area;
				break;
			case VOLUME_ATTRIBUTE:
				attribute = context.volume;
				break;
			default:
			EM_ASM({console.log('Incorrect attribute filter value: using volume');}	);
			attribute = context.volume;
		}
		filterHierarchyAttribute(context, attribute);
	
	
	}
	
    /*if (option.watershed)
    {
        ui("Computing watershed.");
        watershedizing();
        ui("Computing hierarchical watershed.");
        initStructures(size);
        constructBPT(edges, size);
    }
    if (option.attribute != "disabled")
        filterHierarchy(size, option.attribute);

    //removeAttributes();
    ui("Compressing tree.");*/
	
	
	ui("Compressing tree.");
    cleanTree(context);

	ui("Drawing saliency.");
	saliency(context, data);

	
	int nbNodes=context.nbNodes;

	*parent = context.parent;
	*nbnodes = nbNodes;

	deleteGlobalMemory(context);
	delete [] context.depth;
	delete [] context.level;
	
	ui("All done, doing memory stuff :)");
}

}
