#!/usr/bin/env python3

# Copyright ESIEE (2016)
#
# benjamin.perret@esiee.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys
import os
import csv
import numpy
import scipy

def computeStatFolder(folder, imagesIds):
    print(folder)

    with open(imageIds,"r") as idf:
        allids = idf.readlines()

    vals=[]

    for imageid in allids:
        imageid = imageid.strip()
        with open(folder + imageid + ".txt","r") as f:
            line = f.readline()
            fm = float(line.split()[2])
            vals.append(fm)

    vals.sort()
    vals = numpy.array(vals)
    mv = scipy.mean(vals)
    stdv = scipy.std(vals)
    nbElem = len(vals)
    q1 = vals[int(nbElem*0.25)]
    q2 = vals[int(nbElem*0.5)]
    q3 = vals[int(nbElem*0.75)]
    
    IQR = q3 - q1
    outl = q1 - 1.5*IQR
    outu = q3 + 1.5*IQR
    
    lowerI = None
    upperI = None
    for i in range(nbElem):
        if not lowerI and vals[i]>=outl:
            lowerI = i
            break

    for i in range(nbElem-1,-1,-1):
        if not upperI and vals[i]<=outu:
            upperI = i
            break
    
    
    whiskerLow = vals[lowerI]
    whiskerUp = vals[upperI]
    
    outliers = []
    if lowerI != 0:
        outliers.extend(vals[0:lowerI])
    if upperI != nbElem-1:
        outliers.extend(vals[upperI+1:nbElem])
    
    
    
    return [mv, stdv, q1, q2, q3, whiskerLow, whiskerUp, outliers]


if len(sys.argv) != 4:
    raise Exception("Incorrect usage: folder imagesIds outFile")
inFolder = sys.argv[1]
imageIds = sys.argv[2]
outFile = sys.argv[3]

result = computeStatFolder(inFolder, imageIds)

with open(outFile, 'w') as f:
    vals = result[0:len(result)-1]
    line = ["{0:.5f}".format(e) if isinstance(e, float) else str(e) for e in vals]
    f.write("\t".join(line))
    f.write("\n")
    for outlier in result[len(result)-1]:
        f.write("{0:.5f}".format(outlier))
        f.write("\n")
