#!/usr/bin/env python3

# Copyright ESIEE (2016)
#
# benjamin.perret@esiee.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


__author__ = 'perretb'
import os
import csv
import math
import numpy
import scipy
import json
import matplotlib.pyplot as plt
import sys
import collections
from matplotlib.colors import hsv_to_rgb



colors = [[240, 163, 255], [0, 117, 220], [153, 63, 0], [76, 0, 92], [25, 25, 25], [0, 92, 49], [43, 206, 72],
          [255, 204, 153], [128, 128, 128], [148, 255, 181], [143, 124, 0], [157, 204, 0], [194, 0, 136], [0, 51, 128],
          [255, 164, 5], [255, 168, 187], [66, 102, 0], [255, 0, 16], [94, 241, 242], [0, 153, 143], [224, 255, 102],
          [116, 10, 255], [153, 0, 0], [255, 255, 128], [255, 255, 0], [255, 80, 5]]




def clamp(x):
    return max(0, min(x, 255))


def RGBToHex(r, g, b):
    return "#{0:02x}{1:02x}{2:02x}".format(clamp(r), clamp(g), clamp(b))


def formatFloat(f):
    if isinstance(f, collections.Iterable):
        return "(" + ", ".join(("{0:.2f}".format(e) for e in f)) +")"
    else:
        return "{0:.2f}".format(f)


json.encoder.FLOAT_REPR = lambda o: format(o, '.2f')

scoreIndex = 2


def computeStatFolder(folder, maxRegion=99, useMedian=True, indexes=[1]):
    print(folder)
    histo = [None] * maxRegion

    for i in range(len(histo)):
        histo[i] = []
        for j in indexes:
            histo[i].append([])

    for file in os.listdir(folder):
        if file.endswith(".txt"):
            with open(folder + "/" + file, newline='') as csvfile:
                spamreader = csv.reader(csvfile, delimiter='\t')
                spamreader.__next__()  # skip header very dirty
                for row in spamreader:
                    nbR = int(row[0])
                    if 0 < nbR <= maxRegion:
                        binNum = nbR - 1
                        for j in range(len(indexes)):
                            histo[binNum][j].append(float(row[indexes[j]]))

    result = []
    for j in indexes:
        result.append([])
    for i in range(len(histo)):
        for j in range(len(indexes)):
            binvals = numpy.array(histo[i][j])

            if useMedian:
                st = scipy.percentile(binvals, [25, 50, 75])
                newVals = st.tolist()

            else:
                mv = scipy.mean(binvals)
                stdv = scipy.std(binvals)
                newVals = [mv - stdv, mv, mv + stdv]

            result[j].append(newVals)

    return result


def makePlot(xlabels, stats, title, col):
    x = range(len(xlabels))
    y = [v[1] for v in stats]
    lower_error = [(v[1] - v[0]) for v in stats]
    upper_error = [(v[2] - v[1]) for v in stats]
    asymmetric_error = [lower_error, upper_error]

    plt.errorbar(xlabels, y, yerr=asymmetric_error, fmt='-o', label=title, color=col)
    plt.xscale('log')


def process(database, errorMeasure, methodList):
    plt.figure()
    hueStep = 1.0 / (len(methodList) - 1)
    hue = 0
    for method in methodList:
        folder = inFolder + database + "/" + method + "/" + errorMeasure + "/"
        xlabels, stat = computeStatFolder(folder)
        col = hsv_to_rgb(((hue * 1.3) % 1.0, 0.5 + (hue * 7) % 0.5, 1.0))
        makePlot(xlabels, stat, str(method), col)
        hue += hueStep
        # print(xlabels)
    plt.title(database + " - " + errorMeasure)
    plt.ylim([0.0, 1.0])
    # plt.legend(loc=4)
    lgd = plt.legend(loc=9, bbox_to_anchor=(0.5, -0.1), ncol=2)
    # plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
    art = [lgd]
    plt.savefig('result/' + database + "-" + errorMeasure + ".pdf", additional_artists=art, bbox_inches="tight")
    plt.close()


def getIndices():
    op = []
    op.extend(range(10))
    op.extend(range(11, 30, 2))
    op.extend(range(33, 70, 4))
    op.extend(range(74, 99, 5))
    return op

indices = getIndices()


def extractData(tab, ind):
    if len(tab) == 100:
        return [[i + 1, tab[i][ind]] for i in getIndices()]
    return [[int(tab[i][0]), tab[i][ind]] for i in range(len(tab))]
    j = 0
    i = 0
    data = []
    while i != len(indices) and j != len(tab):
        if indices[i] == tab[j][0]:
            data.append([indices[i]+1, tab[j][ind]])
            i += 1
            j += 1
        elif indices[i] < tab[j][0]:
            i += 1
        else:
            j += 1
    return data


def CSVRead(file, header=True, delimiter='\t'):
    data = []

    with open(file, newline='') as csvfile:
        spamreader = csv.reader(csvfile, delimiter=delimiter)
        if header:
            csvfile.readline()
        lines = csvfile.readlines()
        for row in lines:
            try:
                data.append([float(e) for e in row.split()])
            except Exception as e:
                print(e)
                print(row)

    return data

def CSVReadOld(file, header=True, delimiter='\t'):
    data = []

    with open(file, newline='') as csvfile:
        spamreader = csv.reader(csvfile, delimiter=delimiter)
        if header:
            spamreader.__next__()  # skip header very dirty
        for row in spamreader:
            try:
                data.append([float(e) for e in row])
            except:
                print(row)

    return data


def linInterp(data):
    if len(data) == 100:
        return data
    ndata = []
    for i in range(len(data)-1):
        vi = data[i][1]
        ki = data[i][0]
        vii = data[i+1][1]
        kii = data[i+1][0]
        ndata.append([ki, vi])

        for j in range(ki+1, kii):
            ndata.append([j, (vii-vi)/(kii-ki)*(j-ki)+vi])

    ndata.append([data[-1][0], data[-1][1]])
    if len(ndata) != 100:
        for i in range(len(ndata), 100):
            ndata.append([i-1, 0])
    return ndata


def computeKCurveStat(data, dataThreshold, logx=False):
    areaUnderCurve = 0
    areaUnderCurveThreshold = 0
    data = linInterp(data)
    dataThreshold = linInterp(dataThreshold)
    base = data[0][1]
    totalLength = 0
    if logx:
        totalLength = math.log(len(data) + 1)
        for i in range(len(data)):
            interval = math.log(i + 2) - math.log(i + 1)
            areaUnderCurve += data[i][1] * interval
            areaUnderCurveThreshold += dataThreshold[i][1] * interval
    else:
        totalLength = len(data)
        for i in range(len(data)):
            areaUnderCurve += data[i][1]
            areaUnderCurveThreshold += dataThreshold[i][1]
    #normalise = (totalLength * (1.0 - data[0][1]))
    #if normalise == 0:
    #    normalise = 1
    #areaUnderCurve = (areaUnderCurve - base * totalLength) / normalise
    #areaUnderCurveThreshold = (areaUnderCurveThreshold - base * totalLength) / normalise
    areaUnderCurve /= totalLength
    areaUnderCurveThreshold /= totalLength
    return areaUnderCurve, areaUnderCurveThreshold

# depracted median
def processHighChart(database, errorMeasure, methodList, useMedian=False):
    series = []
    count = 0
    indices = getIndices()
    for method in methodList:
       # print("kcurve " + database + " " +  errorMeasure + " " + method)
        coul = RGBToHex(*colors[count % len(colors)])

        if errorMeasure == "UB":
            folder = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + "_UBFB_eval_bdry_thr.txt"
            folderThreshold = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + "_PTFB_eval_bdry_thr.txt"

            header = False
            delimiter = ' '
            ind = 3
        else:
            folder = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + ".txt"
            folderThreshold = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + "_threshold.txt"
            header = True
            delimiter = '\t'
            ind = 1
        if not os.path.isfile(folder) or not os.path.isfile(folderThreshold):
            print("Result not found :" +folder +" " + folderThreshold)
            continue
        stat = CSVRead(folder, header=header, delimiter=delimiter)
        statThreshold = CSVRead(folderThreshold)
        data = extractData(stat, ind) #[[i + 1, stat[i][4]] for i in indices]
        dataThreshold = extractData(statThreshold, ind) #[[i + 1, statThreshold[i][4]] for i in indices]
        # if useMedian:
        #     data = extractData(stat, 4) #[[i + 1, stat[i][4]] for i in indices]
        #     dataThreshold = extractData(statThreshold, 4) #[[i + 1, statThreshold[i][4]] for i in indices]
        # else:
        #     data = extractData(stat, 1) #[[i + 1, stat[i][1]] for i in indices]
        #     dataThreshold = extractData(statThreshold, 1) #[[i + 1, statThreshold[i][1]] for i in indices]

        areaUnderCurve, areaUnderCurveThreshold = computeKCurveStat(data, dataThreshold, logx=True)

        serie = {
            "id": method,
            "name": method + "(AUC: " + formatFloat(areaUnderCurve) + ", AUCT: " + formatFloat(
                areaUnderCurveThreshold) + ")",
            "visible": count < 4,
            "data": data,
            "color": coul
        }
        series.append(serie)

        sThreshold = {
            "name": method + ' Threshold',
            "type": 'line',
            "data": dataThreshold,
            "linkedTo": method,
            "color": coul,
            "dashStyle": 'dash'
        }
        series.append(sThreshold)
        # statMarkerNegatif = stat[1]
        # range = [[i+1, statMarkerNegatif[i][3]] for i in indices]
        # sRange = {
        #     "name": method + ' Range',
        #     "type": 'line',
        #     "data": range,
        #     "linkedTo": method,
        #     "color": coul,
        #     "yAxis": 1,
        #     "dashStyle": 'dash'
        # }
        # series.append(sRange)
        if False:
            if useMedian:
                dataRange = [[i + 1, stat[i][3], stat[i][5]] for i in indices]
            else:
                dataRange = [[i + 1, stat[i][1] - stat[i][2], stat[i][1] + stat[i][2]] for i in indices]
            srange = {
                "name": method + ' Dispersion',
                "data": dataRange,
                "type": 'arearange',
                "lineWidth": 0,
                "linkedTo": method,  # ':previous' ,
                "color": coul,
                "fillOpacity": 0.2,
                "zIndex": 0
            }
            series.append(srange)
        count += 1

    plot1 = {
        "title": {
            "text": database + ' ' + errorMeasure
        },

        "chart": {
            "ignoreHiddenSeries": False,
        },

        "xAxis": {
            "type": 'logarithmic',
            "minorTickInterval": 0.1,
            "title": {
                "text": "Number of regions"
            }
        },
        "yAxis": [{
            "min": 0,
            "max": 1,
            "title": {
                "text": "Score"
            }
        },
            #     {
            #     "type": 'logarithmic',
            #     "title": {
            #         "text": "Range",
            #         },
            #     "opposite": True
            # }
        ],

        "series": series
    }
    return plot1


def processKCurveJSON(inFolder, database, errorMeasure, methodList, normalize=True):
    series = []
    count = 0
    indices = getIndices()
    for method in methodList:

       # print("kcurve " + database + " " +  errorMeasure + " " + method)
        coul = RGBToHex(*colors[count % len(colors)])


        folder = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + ".txt"
        folderThreshold = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + "_threshold.txt"

        if not os.path.isfile(folder) or not os.path.isfile(folderThreshold):
            print("Result not found :" +folder +" " + folderThreshold)
            continue
        with open(folder, 'r') as infile:
            data = json.load(infile)

        with open(folderThreshold, 'r') as infile:
            dataT = json.load(infile)

        if normalize:
            data_lin = data["curve_norm"]
            data_ods_auc_lin = data["ods_auc_norm"]
            data_ods_best_lin = data["ods_best_norm"]
            data_ois_best_lin = data["ois_best_norm"]

            dataT_lin = dataT["curve_norm"]
            dataT_ods_auc_lin = dataT["ods_auc_norm"]
            dataT_ods_best_lin = dataT["ods_best_norm"]
            dataT_ois_best_lin = dataT["ois_best_norm"]
        else:

            data_lin = data["curve_lin"]
            data_ods_auc_lin = data["ods_auc_lin"]
            data_ods_best_lin = data["ods_best_lin"]
            data_ois_best_lin = data["ois_best_lin"]

            dataT_lin = dataT["curve_lin"]
            dataT_ods_auc_lin = dataT["ods_auc_lin"]
            dataT_ods_best_lin = dataT["ods_best_lin"]
            dataT_ois_best_lin = dataT["ois_best_lin"]



        serie = {
            "id": method,
            "name": method + "(AUC: " + formatFloat(data_ods_auc_lin) + ", AUCT: " + formatFloat(
                dataT_ods_auc_lin) + ")",
            "visible": count < 4,
            "data": data_lin,
            "color": coul
        }
        series.append(serie)

        sThreshold = {
            "name": method + ' Threshold',
            "type": 'line',
            "data": dataT_lin,
            "linkedTo": method,
            "color": coul,
            "dashStyle": 'dash'
        }
        series.append(sThreshold)



        count += 1

    plot1 = {
        "title": {
            "text": database + ' ' + errorMeasure + (" Fragmentation curve" if normalize else "")
        },

        "chart": {
            "ignoreHiddenSeries": False,
        },

        "xAxis": {
            "type": 'linear' if normalize else 'logarithmic',
            "minorTickInterval": 0.1,
            "title": {
                "text": "Number of regions"
            }
        },
        "yAxis": [{
            "min": 0,
            "max": 1,
            "title": {
                "text": "Score"
            }
        },
            #     {
            #     "type": 'logarithmic',
            #     "title": {
            #         "text": "Range",
            #         },
            #     "opposite": True
            # }
        ],

        "series": series
    }
    return plot1

def processKRepartition(inFolder, database, errorMeasure, methodList, threshold):
    series = []
    count = 0

    maxV = 0
    for method in methodList:

        coul = RGBToHex(*colors[count % len(colors)])

        if threshold:
            folder = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + "_threshold.txt"
        else:
            folder = inFolder + method + os.path.sep + method + "_" + database + "_" + errorMeasure + ".txt"


        if not os.path.isfile(folder) :
            print("Result not found :" +folder )
            continue
        with open(folder, 'r') as infile:
            data = json.load(infile)

        xx = []
        yy = []


        resDataset = data["dataset"]
        for e in resDataset.values():
            el = e["score_ogt"]
            for t in el:
                xx.append(t[0])
                yy.append(t[1])
        maxV = max(maxV,max(xx),max(yy))
        data = list(zip(xx, yy))

        serie = {
            "id": method,
            "name": method,
            "visible": count < 4,
            "data": data,
            "color": coul
        }
        series.append(serie)


        count += 1


    op = {
        "name": 'y=x',
        "type": 'line',
        "data": ((1,1),(maxV,maxV)),
        "color": "#BBBBBB",
        "visible": True,
        "lineWidth": 1,
        "showInLegend": False,
        "marker": {
            "fillColor": 'transparent'
        },
    }
    series.append(op)
    plot1 = {
        "title": {
            "text": database + ' ' + errorMeasure
        },

        "chart": {
            "type": "scatter",
            "ignoreHiddenSeries": False,
        },

        "xAxis": {
            "type": 'linear',
            "title": {
                "text": "Number of regions in GT"
            }
        },
        "yAxis": [{
            "title": {
                "text": "Number of regions for best Score"
            }
        },

        ],

        "series": series
    }
    return plot1



def frange(x, y, jump):
    while x < y:
        yield x
        x += jump


def genISOCurveFMeasure():
    curves = []
    for fmi in range(1, 10):
        fm = fmi / 10.0
        c = []
        prmin = fm / 2.0 + 0.001
        for pr in frange(prmin, 1.00001, 0.01):
            re = (pr * fm) / (2 * pr - fm)
            c.append([pr, re])
        curves.append([fm, c])

    return curves


def processMarkerChart(inFolder, database, marker, methodList):
    series = []
    count = 0
    for method in methodList:

        coul = RGBToHex(*colors[count % len(colors)])

        folder = inFolder + method + os.path.sep + method + "_" + database + "_"
        file = folder + "MARKER_" + marker + ".txt"
        if not os.path.isfile(file): # backward compat
            file = folder + marker + ".txt"
        if os.path.isfile(file):
            outliers = []
            with open(file) as f:
                line = f.readline()
                tokens = line.split()
                mean = float(tokens[0])
                stddev = float(tokens[1])
                q1 = float(tokens[2])
                q2 = float(tokens[3])
                q3 = float(tokens[4])
                whiskerLow = float(tokens[5])
                whiskerUp = float(tokens[6])
                lines = f.readlines()
                for line in lines:
                    outliers.append(float(line))

            serie = {
                "id": method,
                "name": method,
                "visible": count < 4,
                "type": "boxplot",
                "data": [[whiskerLow, q1, q2, q3, whiskerUp]],
                "color": coul,
            }
            series.append(serie)

            count += 1
        else:
            print("File not found: " + file)
    plot1 = {
        "title": {
            "text": database + ' ' + marker
        },

        "chart": {
            "inverted": True
        },
        "yAxis": {
            "min": 0,
            "max": 1,
            "title": {
                "text": "F-Measure"
            }
        },
        "xAxis": {
            "title": {
                "text": ""
            }
        },

        "series": series
    }
    return plot1


def processAllPartition(inFolder, outFolder, methods):
    charts = []


    group = []
    for database in ["bsds500"]:
        for measure in ["BCE"]:


            group.append({
                "title": database + ' - ' + measure + ' (normalized k))',
                "chart": processKCurveJSON(inFolder, database, measure, methods, True)})


    charts.append(group)

    group = []
    for database in ["bsds500"]: 
        for measure in ["BCE"]:
            group.append({
                "title": database + ' - ' + measure + ' (optimal))',
                "chart": processKRepartition(inFolder, database, measure, methods, False) })


    charts.append(group)




    with open(outFolder + 'data2.js', 'w') as f:
        f.write("chartlist = JSON.parse('")
        json.dump(charts, f)
        f.write("');\n")




def processMarkerMeasures(inFolder, outFolder, methods, databases=["grabcut", "weizmann1obj"],
                          markers=["markerll", "markerfs", "markerss"]):
    charts = []
    for database in databases:
        group = []

        for marker in markers:
            group.append({
                "title": marker,
                "chart": processMarkerChart(inFolder, database, marker, methods)})
        charts.append(group)

    with open(outFolder + 'dataMarker.js', 'w') as f:
        f.write("chartlist = JSON.parse('")
        json.dump(charts, f)
        f.write("');\n")


def findResults(inFolder):

    allMethods = [m for m in os.listdir(inFolder) if os.path.isdir(inFolder+m)]
    allMethods.sort()
    return allMethods


def main():
    
    inFolder = "./results/"
    outFolder = "./webPlot/data/"

    print("Processing folder: "  + inFolder)
    allMethods=findResults(inFolder)
    
    print(allMethods)
    processAllPartition(inFolder, outFolder, allMethods)
    processMarkerMeasures(inFolder, outFolder, allMethods) 

if __name__ == '__main__':
    main()
